from enum import Enum

import requests

from .annotate import Annotation, Annotator, Corpus, Document, Task


def json2annotator(js):
    return Annotator(
        name=js['name'],
        annotator_id=js['id'],
        model=js['model']
    )


def json2annotation(js):
    return Annotation(
            entity_id=js['entity'],
            body=js['body'],
            annotator=js['annotator'],
            document_id=js['document'],
            task_id=js['task'],
            target=js['target'],
            created=js['created']
        )


def json2doc(js):
    return Document(
        uri=js['uri'],
        content=js['content'],
        corpus=js['corpus'],
        document_id=js['id']
    )


def json2task(js):
    return Task(
        name=js['name'],
        description=js['description'],
        entities=js['entities'],
        corpora=js['corpora'],
        annotators=js['annotators'],
        task_id=js['id']
    )


class AssignmentType(Enum):
    REVIEW = 'R'
    LABEL = 'A'


class AssignmentStatus(Enum):
    ASSIGNED = 'A'
    COMPLETED = 'C'


class LinalgoClient:

    endpoints = {
        'annotators': 'annotators',
        'annotations': 'annotations',
        'corpora': 'corpora',
        'documents': 'documents',
        'entities': 'entities',
        'task': 'tasks',
    }

    def __init__(self, token, api_url="http://localhost:8000"):
        self.api_url = api_url
        self.access_token = token

    def request(self, url, query_params={}):
        headers = {'Authorization': f"Token {self.access_token}"}
        res = requests.get(url, headers=headers, params=query_params)
        if res.status_code == 401:
            raise Exception(f"Authentication failed. Please check your token.")
        if res.status_code == 404:
            raise Exception(f"{url} not found.")
        elif res.status_code != 200:
            raise Exception(f"Request returned status {res.status_code}")
        return res.json()

    def get_corpora(self):
        res = self.request(self.endpoints['corpora'])
        corpora = []
        for js in res['results']:
            corpus_id = js['id']
            corpus = self.get_corpus(corpus_id)
            corpora.append(corpus)
        return corpora

    def get_corpus(self, corpus_id):
        url = f"{self.enpoints['corpora']}/{corpus_id}/"
        res = self.request(url)
        corpus = Corpus(name=res['name'], description=res['description'])
        documents = self.get_corpus_documents(corpus_id)
        corpus.documents = documents
        return corpus

    def get_corpus_documents(self, corpus_id):
        url = f"/corpora/{corpus_id}/documents/?page_size=1000"
        res = self.request(url)
        documents = []
        for js in res['results']:
            document = json2doc(js)
            documents.append(document)
        return documents

    def get_tasks(self, task_ids=[]):
        url = "tasks/"
        tasks = []
        res = self.request(url)
        if len(task_ids) == 0:
            for js in res['results']:
                task_ids.append(js['id'])
        for task_id in task_ids:
            task = self.get_task(task_id)
            tasks.extend(task)
        return tasks

    def get_task_documents(self, task_id):
        query_params = {'corpus__tasks': task_id, 'page_size': 1000}
        docs = []
        next_url = "{}/{}/".format(
            self.api_url, self.endpoints['documents'])
        while next_url:
            res = self.request(next_url, query_params)
            next_url = res['next']
            docs.extend(res['results'])
        return [json2doc(doc_json) for doc_json in docs]

    def get_task_annotations(self, task_id):
        query_params = {'task': task_id, 'page_size': 1000}
        docs = []
        next_url = "{}/{}/".format(self.api_url, self.endpoints['annotations'])
        while next_url:
            res = self.request(next_url, query_params)
            next_url = res['next']
            docs.extend(res['results'])
        return [json2annotation(a_json) for a_json in docs]

    def get_task(self, task_id, verbose=False):
        task_url = "{}/{}/{}/".format(
            self.api_url, self.endpoints['task'], task_id)
        if verbose:
            print(f'Retrivieving task with id {task_id}...')
        task_json = self.request(task_url)
        task = json2task(task_json)
        if verbose:
            print('Retrieving annotators...')
        task.annotators = self.get_annotators(task)
        if verbose:
            print('Retrieving entities...')
        params = {'tasks': task.id, 'page_size': 1000}
        entities_url = "{}/{}".format(self.api_url, self.endpoints['entities'])
        entities_json = self.request(entities_url, params)
        task.entities = entities_json['results']
        if verbose:
            print('Retrieving documents...')
        task.documents = self.get_task_documents(task_id)
        if verbose:
            print('Retrieving annotations...')
        task.annotations = self.get_task_annotations(task_id)
        return task

    def get_annotators(self, task=None):
        params = {'tasks': task.id, 'page_size': 1000}
        annotators_url = "{}/{}/".format(
            self.api_url, self.endpoints['annotators'])
        res = self.request(annotators_url, params)
        annotators = []
        for js in res['results']:
            annotator = json2annotator(js)
            annotators.append(annotator)
        return annotators

    def create_annotator(self, annotator):
        if annotator.annotator_id is not None:
            raise Exception("Annotator already has an ID.")
        annotator_url = "{}/{}/".format(
            self.api_url, self.endpoints['annotators'])
        url = self.api_url + annotator_url
        headers = { 'Authorization': f"Token {self.access_token}"}
        annotator_json = {
            'name': annotator.name,
            'model': str(annotator.model)
        }
        res = requests.post(url, json=annotator_json, headers=headers).json()
        annotator.annotator_id = res['id']
        annotator.owner = res['owner']
        return annotator

    def create_annotations(self, annotations):
        url = "{}/{}/".format(self.api_url, self.endpoints['annotations'])
        headers = {'Authorization': f"Token {self.access_token}"}
        res = requests.post(url, json=annotations, headers=headers)
        return res

    def assign(self, document, annotator, task, reviewee=None,
               assignment_type=AssignmentType.LABEL.value):
        doc_status = {
            'status': AssignmentStatus.ASSIGNED.value,
            'type': assignment_type,
            'document': document,
            'annotator': annotator,
            'task': task,
            'reviewee': reviewee
        }
        url = self.api_url + '/document-status/'
        headers = {'Authorization': f"Token {self.access_token}"}
        res = requests.post(url, doc_status, headers=headers)
        return res

    def unassign(self, status_id):
        headers = {'Authorization': f"Token {self.access_token}"}
        url = "{}/{}/{}/".format(self.api_url, '/document-status/', status_id)
        res = requests.delete(url, headers=headers)
        return res

    def get_schedule(self, task):
        query_params = {'task': task.id, 'page_size': 1000}
        docs = []
        next_url = "{}/{}/".format(self.api_url, '/document-status/')
        while next_url:
            res = self.request(next_url, query_params=query_params)
            next_url = res['next']
            docs.extend(res['results'])
        return docs



import numpy as np
import pandas as pd

from .client import AssignmentStatus


class AnnotatorNotFound(Exception):
    pass


class NotEnoughReviews(Exception):
    pass


class Scheduler:

    def __init__(self, task, schedule):
        self.task = task
        self.schedule = schedule

    def unseen_documents(self, n):
        """
        Parameters
        ----------

        n: int
            Number of unseen documents to return
        """
        annotated_docs = set(
            annotation.to_json()['document'] for annotation in self.task.annotations)
        docs = set(doc.id for doc in self.task.documents)
        new_docs = list(docs - annotated_docs)

        if len(new_docs) < n:
            raise NotEnoughReviews()

        return set(np.random.choice(new_docs, size=n, replace=False))


    def random_review(self, reviewer_id, reviewee_id, n=None):
        """
        Parameters
        ----------
        reviewer_id: uuid
            The uuid of the reviewer
        reviewee_id: uuid
            The uuid of the reviewee
        n: int
            The number of documents to review

        Return
        ------
        A set of documents to review
        """

        if (reviewer_id not in {a.id for a in self.task.annotators} or
            reviewee_id not in {a.id for a in self.task.annotators}):
            raise AnnotatorNotFound()
        reviewer_idx = self.schedule['annotator'] == reviewer_id
        reviewee_idx = self.schedule['annotator'] == reviewee_id
        seen_idx = self.schedule['status'] == AssignmentStatus.COMPLETED.value
        pending_idx = self.schedule['status'] == AssignmentStatus.ASSIGNED.value

        idx = reviewer_idx & (seen_idx | pending_idx)
        reviewer_docs = set(self.schedule.loc[idx, 'document'])
        idx = reviewee_idx & seen_idx
        reviewee_docs = set(self.schedule.loc[idx, 'document'])

        pool = list(reviewee_docs - reviewer_docs)
        if n is not None:
            if n > len(pool):
                print(pool)
                raise NotEnoughReviews()
            return set(np.random.choice(pool, size=n, replace=False))
        return pool

    def random_assign(self, assignee_id, n):
        """

        Parameters
        ----------

        assignee_id: uuid
            The uuid of the annotator
        n: int
            the number of documents

        Return
        ------
        A set of documents to assign
        """

        if assignee_id not in {a.id for a in self.task.annotators}:
            raise AnnotatorNotFound(
                '{} is not a known annotator'.format(assignee_id))

        all_docs = set(doc.id for doc in self.task.documents)
        all_seen_docs = set()
        for annotation in self.task.annotations:
            all_seen_docs.add(annotation.to_json()['document'])
        all_new_docs = all_docs - all_seen_docs

        assignee_idx = self.schedule['annotator'] == assignee_id
        seen_idx = self.schedule['status'] == AssignmentStatus.COMPLETED.value
        pending_idx = self.schedule['status'] == AssignmentStatus.ASSIGNED.value

        idx = assignee_idx & (seen_idx | pending_idx)
        assignee_docs = set(self.schedule.loc[idx, 'document'])

        pool = list(all_new_docs - assignee_docs)
        if n > len(pool):
            print(pool)
            raise NotEnoughReviews()
        return set(np.random.choice(pool, size=n, replace=False))




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix


def plot_confusion_matrix(y_true, y_pred, classes,
                          normalize=False,
                          title=None,
                          names=None,
                          cmap=plt.cm.Blues,
                          ax=None):

    # Compute confusion matrix
    cm = confusion_matrix(y_true, y_pred)
    # Only use the labels that appear in the data
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    if ax is None:
        fig, ax = plt.subplots()
    im = ax.imshow(cm, interpolation='nearest', cmap=cmap)

    ax.set(xticks=np.arange(cm.shape[1]),
           yticks=np.arange(cm.shape[0]),
           # ... and label them with the respective list entries
           xticklabels=classes, yticklabels=classes,
           title=title,
           ylabel=names[0],
           xlabel=names[1])

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(j, i, format(cm[i, j], fmt),
                    ha="center", va="center",
                    color="white" if cm[i, j] > thresh else "black")
    return ax


def plot_matrix(cm, xlabels=None, ylabels=None, title=None, cmap=plt.cm.Blues):
    fig, ax = plt.subplots()
    im = ax.imshow(cm, interpolation='nearest', cmap=cmap)

    ax.set(
        xticks=np.arange(cm.shape[1]),
        yticks=np.arange(cm.shape[0]),
        xticklabels=xlabels,
        yticklabels=ylabels,
        title=title
    )

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    fmt = '.2f'
    thresh = cm.max() / 2.
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(j, i, format(cm[i, j], fmt),
                    ha="center", va="center",
                    color="white" if cm[i, j] > thresh else "black")
    fig.tight_layout()
    return ax


def multiclass_dataframe(task):
    df = []
    for annotation in task.annotations:
        data = {
            'document': annotation.document_id,
            'label': annotation.type_id,
            'annotator': annotation.annotator,
            'created': annotation.created
        }
        df.append(data)
    df = pd.DataFrame(df)
    df['created'] = pd.to_datetime(df['created'])
    df2 = df.loc[df.groupby(['annotator', 'document']).created.idxmax()]
    df3 = df2.pivot(index='document', columns='annotator',
                    values='label').reset_index()
    annotators = [x for x in df3.columns[1:]]
    df3.columns = ['document_id'] + annotators
    df3[annotators] = df3[annotators].applymap(np.vectorize(task.get_name))

    dd = []
    for doc in task.documents:
        data = {
            'document_id': doc.id,
            'content': doc.content
        }
        dd.append(data)
    dd = pd.DataFrame(dd)
    return pd.merge(df3, dd, on='document_id')



from collections import defaultdict, MutableSequence
import uuid


class Annotation:
    """
    Annotation class compatible with the W3C annotation data model.
    """

    def __init__(self, entity_id, body, annotator=None, task_id=None,
                 score=None, document_id=None, annotation_id=None,
                 target={}, created=None):
        self.id = annotation_id
        if self.id is None:
            self.id = uuid.uuid4()
        self.type_id = entity_id
        self.score = score
        self.body = body
        self.task_id = task_id
        self.annotator = annotator
        self.document_id = document_id
        self.target = target
        self.created = created

    def to_json(self):
        js = {
            'id': str(self.id),
            'entity': self.type_id,
            # 'uri': self.uri,
            'type': self.type_id,
            # 'text': self.text,
            'task': self.task_id,
            'annotator': self.annotator,
            'document': self.document_id,
            'target': self.target,
            }
        return js

    def __repr__(self):
        return str(self.type_id)


class Annotations(MutableSequence):
    """
    Class to list of annotations with custom indexing.
    """

    def __init__(self, annotations=None):
        super(Annotations, self).__init__()
        self._list = []
        self._user_index = defaultdict(list)
        self._type_index = defaultdict(list)
        for annotation in annotations:
            self.append(annotation)

    def get_users(self):
        return self._user_index.keys()

    def by_user(self, name):
        return self._user_index[name]

    def __repr__(self):
        return "<{0} {1}>".format(self.__class__.__name__, self._list)

    def __len__(self):
        return len(self._list)

    def __getitem__(self, ii):
        return self._list[ii]

    def __delitem__(self, ii):
        del self._list[ii]

    def __setitem__(self, ii, val):
        self._list[ii] = val

    def __str__(self):
        return str(self._list)

    def insert(self, ii, val):
        self._list.insert(ii, val)

    def append(self, val):
        self.insert(len(self._list), val)
        self._user_index[val.annotator].append(val)
        self._type_index[val.type_id].append(val)


class Annotator:
    """
    The Annotator class can create, delete or modify Annotations.
    """

    def __init__(self, name, task=None, model=None, annotation_type_id=None,
                 threshold=0, annotator_id=None):
        self.id = annotator_id
        self.name = name
        self.task = task
        self.model = model
        self.type_id = annotation_type_id
        self.threshold = threshold
        self.annotator_id = None

    def assign_task(self, task):
        self.task = task

    def _get_annotation(self, document):
        score = self.model.decision_function([document.content])[0]
        if score >= self.threshold:
            label = self.type_id
        else:
            label = 1  # Viewed
        annotation = Annotation(
            uri=f'/tasks/{self.task.id}/annotate/{document.id}',
            type_id=label,
            score=score,
            text=document.content,
            annotator=self.annotator_id,
            task_id=self.task.id,
            document_id=document.id
        )
        return annotation

    def annotate(self, document):
        annotation = self._get_annotation(document)
        if annotation is not None:
            self.task.annotations.append(annotation)
            document.annotations.append(annotation)
        return annotation


class Corpus:

    def __init__(self, name, description, documents=[]):
        self.name = name
        self.description = description
        self.documents = documents


class Document:
    """
    Base class that holds the document on which to perform annotations.
    """

    def __init__(self, uri, content, corpus=None, document_id=None):
        self.uri = uri
        self.content = content
        self.corpus = corpus
        self.id = document_id
        self._annotations = Annotations([])

    @property
    def labels(self):
        return set(annotation.type_id for annotation in self.annotations)

    @property
    def annotations(self):
        return self._annotations

    @annotations.setter
    def annotations(self, values):
        self._annotations = Annotations(values)


class Entity:

    def __initi__(self, name, entity_id=None):
        self.id = entity_id
        self.name = name


class Task:
    """
    The Task class contains all information about a task: entities, corpora, 
    annotations.
    """

    def __init__(self, name, description, entities=None, corpora=None,
                 annotators=None, annotations=Annotations([]), documents=[],
                 task_id=None):
        self.id = task_id
        self.name = name
        self.description = description
        self.entities = entities
        self.corpora = corpora
        self.annotators = annotators
        self._annotations = annotations
        self._documents = documents

    @property
    def documents(self):
        return list(self._documents.values())

    @documents.setter
    def documents(self, values):
        self._documents = dict((doc.id, doc) for doc in values)

    @property
    def annotations(self):
        return self._annotations

    @annotations.setter
    def annotations(self, values):
        self._annotations = Annotations(values)
        for annotation in self.annotations:
            doc_id = annotation.document_id
            if doc_id in self._documents:
                self._documents[doc_id].annotations.append(annotation)
            else:
                # TODO: Investigate the reason for this
                pass

    def transform(self, target='binary', label=None):
        docs = []
        labels = []
        if target == 'binary':
            for doc in self.documents:
                if len(doc.annotations) > 0:
                    docs.append(doc.content)
                    labels.append(1 if label in doc.labels else 0)
            return docs, labels
        elif target == 'multilabel':
            entities = {e['id']: e['title'] for e in self.entities}
            for document in self.documents:
                if len(document.annotations) > 0:
                    docs.append(document.content)
                    labels.append({entities[l]: 1 for l in document.labels})
            return docs, labels
        else:
            # TODO: raise proper exception
            raise Exception('target should be `binary` or `multiclass`')

    def get_name(self, some_id):
        for a in self.annotators:
            if a.id == some_id:
                return a.name

        for e in self.entities:
            if e['id'] == some_id:
                return e['title']
        return some_id

    def get_id(self, name):
        for a in self.annotators:
            if a.name == name:
                return a.id

        for e in self.entities:
            if e['title'] == name:
                return e['id']
        return name

    def __repr__(self):
        rep = (f"name: {self.name}\ndescription: {self.description}\n# "
               f"documents: {len(self.documents)}\n# annotations: "
               f"{len(self.annotations)}")
        if self.id:
            rep = f"id: {self.id}\n" + rep
        return rep



from setuptools import setup

setup(
    name='linalgo',
    packages=['linalgo'],
    description=('Python library for Natural Language Processing supporting '
                 'the open annotation standard'),
    version='0.0',
    author='Arnaud Rachez',
    author_email='arnaud@linalgo.com',
    download_url = 'https://github.com/DiscoverSDK/pysimplib/tarball/0.1',
    requires=['numpy', 'scipy'],
)



import string
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from itertools import accumulate

from sklearn.metrics import confusion_matrix


def tokenize(documents, orient='dict'):
    """
    Transform documents into a collection of token

    Parameters
    -------------
    documents: list[Documents]
        A list of document objects containing
    orient: str, {'dict', 'record'}
        The  format of the returned dictionary

    Returns
    ---------
    Dict[str, Dict]
    """
    tok_map = {}
    for doc in documents:
        tokens = doc.content.split()
        end = list([e + i - 1 for i, e in
                    enumerate(accumulate([len(tok) for tok in tokens]))])
        start = [s + 2 for s in [-2] + end[:-1]]
        c = {'token': tokens, 'start': start, 'end': end}
        if orient == 'dict':
            tok_map[doc.id] = c
        elif orient == 'record':
            tok_map[doc.id] = [{'token': t, 'start': s, 'end': e} for t, s, e in
                               zip(tokens, start, end)]
        else:
            raise Exception("`orient` should be in {'dict', 'record'}")
    return tok_map


def is_punct(x):
    if x in string.punctuation:
        return True
    if x in ['-RRB-', '-LRB-']:
        return True
    return False


def compare_tags(task, untag_punct=True, min_annotators=2):
    """
    Calculate tag alignment for several annotators on the same documents

    Parameters
    -------------
    task : Task
        The task object to compute alignement from
    untag_punct : bool
        Whether or not to automatically untag punctuation tokens

    Returns
    ---------
    Dict
        Records containing the token and associated tags for each annotator
    """
    entities = {e['id']: e['title'] for e in task.entities}
    tok_map = tokenize(task.documents, orient='record')
    al = []
    for doc in task.documents:
        tokens = tok_map[doc.id]
        xl = pd.DataFrame(tokens)
        annotators = list({a.annotator for a in doc.annotations})
        if len(annotators) < min_annotators:
            continue
        for annotator in annotators:
            xl[annotator] = 'O'
        for anno in doc.annotations:
            start = anno.target['selector'][0]['startOffset']
            end = anno.target['selector'][0]['endOffset']
            idx1 = xl['start'] >= start
            idx2 = xl['end'] <= end
            xl.loc[idx1 & idx2, anno.annotator] = entities[anno.type_id]
        if untag_punct:
            xl.loc[xl['token'].apply(is_punct), annotators] = 'O'
        al.append(xl.to_dict(orient='record'))
    return al


def filter_by_entity(al, entity, annotators):
    xl = pd.DataFrame(al)
    idx = np.array([False ]* xl.shape[0])
    for a in annotators:
        if a in xl.columns:
            idx = idx | (xl[a] == entity)
    return xl[idx]


def score(al, entity, annotators):
    df = filter_by_entity(al, entity['title'], annotators)
    annotators = [a for a in annotators if a in df.columns]
    return sum(df[annotators[0]] == df[annotators[1]]) / len(df)


def plot_confusion_matrix(
        als, task, normalize=False, title=None, cmap=plt.cm.Blues):
    xls = []
    for al in als:
        xl = pd.DataFrame(al)
        xl = xl[[a for a in xl.columns if a in task.annotators]]
        xl = xl.iloc[:, [0, 1]]
        xl.columns = ['a', 'b']
        xls.append(xl)

    cc = pd.concat(xls)
    classes = sorted(list(cc['b'].unique()))
    y_true = cc.values[:, 1]
    y_pred = cc.values[:, 0]

    if not title:
        if normalize:
            title = 'Normalized confusion matrix'
        else:
            title = 'Confusion matrix, without normalization'

    # Compute confusion matrix
    cm = confusion_matrix(y_true, y_pred)
    # Only use the labels that appear in the data
#     classes = classes[unique_labels(y_true, y_pred)]
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    fig, ax = plt.subplots()
    im = ax.imshow(cm, interpolation='nearest', cmap=cmap)
    # ax.figure.colorbar(im, ax=ax)
    # We want to show all ticks...
    ax.set(xticks=np.arange(cm.shape[1]),
           yticks=np.arange(cm.shape[0]),
           # ... and label them with the respective list entries
           xticklabels=classes, yticklabels=classes,
           title=title,
           ylabel='Annotator A',
           xlabel='Annotator B')

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    fmt = '.2f' if normalize else 'd'
    thresh = cm.max() / 2.
    for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
            ax.text(j, i, format(cm[i, j], fmt),
                    ha="center", va="center",
                    color="white" if cm[i, j] > thresh else "black")
    fig.tight_layout()
    return ax
